/*               Griddy-Gibbs for GARCH-Student - Simulated data 

This program is needed to reproduce the results for Tables 1 (all results), 2, 3 and 4 
(Griddy-Gibbs results) of the paper:
L. Bauwens and M. Lubrano "Bayesian inference on GARCH models using the Gibbs sampler"

This version of the program generates the results in Table 4 (Griddy-Gibbs column), ie for 
-a flat prior on nu on (0,30)
-intervals of integration are assigned in the TETA_INF and TETA_SUP variables
-2000 total draws of which the initial 1000 draws are discarded
(see N = 2000 and NDROP = 1000 instructions on lines 155-156)

This program must be slightly changed to reproduce other results:
-for a half-Cauchy prior on nu: see PROC POGARCH (select the appropriate RETP instruction, 
 see lines 291-296). Also see lines 224-227 to write the selected prior in the output.
-adjust the upper limit of integration of nu (set last element of TETA_SUP equal to 18 
  instead of 30 in the present version), see lines 128-129.
-if you wish (strongly advised): change the output file names TIR150AF, TES150AF, TVA150AF, 
 TVA150AF, and MAR150AF to whatever you like (see lines 202-208). This is useful 
 to keep the draws and the results for each case run. There is a follow-up program called 
 GGF.PRG which uses these files as input in order to compute more results, to draw graphs 
 of marginal densities, and to compute, print and plot the CUMSUM statistics (see comments 
 in the heading of that program). Also change the output file name gg150AF (line 32), and 
 the SPATH instruction (line 35).

Of course, there are many other options which can be changed: see comments in the program.
This program does also the maximum likelihood estimation (using PROC LFGARCH)
*/ 


library pgraph,maxlik;
OUTPUT file = c:\mydoc\articles\gibbs\calculs\gg150AF.ou1 reset;
graphset;
fonts ("simplex complex microb simgrma");
SPATH="c:\\mydoc\\articles\\gibbs\\calculs"; @path for saving results@
datestr(0);;" at ";;timestr(0);
__title = "Artificial data (Student) - Griddy Gibbs sampling";
print $__title;
format 8,4;


@@@ Generate data: DGP is GARCH(1,1) @@@

rndseed 13579;
nt = 151;
yt = zeros(nt,1);
alpha = 0.4;
beta =  0.4;
nu = 5;
omega = 0.1;

y = ones(nt,1);
u = zeros(nt,1);
ht = ones(nt,1);
dep = 1; @starting value for evaluation of ht in l.f.@

i = 2; do while i<=nt;
    ht[i] = omega + alpha*u[i-1]^2 + beta*ht[i-1];
    u[i] = studt(0,sqrt(ht[i]),nu,1);
    y[i] = u[i];
i = i+1; endo;

"omega = ";; omega;;"alpha = ";;alpha;;"   beta = ";;beta;;"   nu = ";;nu;
@"rho = ";;rho;;"   mu = ";;mu;@
if nu > 2;
   wsc = 1 - alpha*nu/(nu-2) - beta;
   "weak stationarity condition ";;wsc;
   if wsc > 0;
      uv = omega/wsc;
      "unconditional variance ";; uv;
   else;
      "unconditional variance does not exist";
   endif;
else;
   "nu smaller than 2";
endif;
@
var = cons;
xy(seqa(1,1,nt),y~sqrt(ht));
@

stdy = stdc(y);
"EcT initial de la serie = ";;stdy;
y = y/stdy;
dep = dep/stdy;

yt = y[2:nt];
y1 = y[1:nt-1];
x = yt;
uq = yt^2;
t = rows(yt);
vy = vcx(yt);
@save path=^spath y;@
@"y";;y;@
"";"number of observations";;t;
"mean and s.d.";;meanc(y)~sqrt(vy);


@@@ Maximum likelihood estimation with MAXLIK @@@

#include maxlik.ext;
maxset;
__output = 50;
_mlparnm = "o_root" | "a_root" | "g_root" | "nu   ";
@ GARCH parameters are square roots of original ones, see lfgarch: @
teta_i = sqrt(omega)|sqrt(alpha)|sqrt(beta)|nu;
@starting values for m@
"";"teta_i";;teta_i';
{teta_h,l,s,cov,retcode} = maxprt(maxlik(yt,0,&lfgarch,teta_i));
teta_ho = teta_h[1:3]^2 | teta_h[4] ;
save path=^spath teta_ho;
jc = zeros(4,4);
jd =  2*teta_h[1:3] | 1; @jacobian (diagonal)@
jc = diagrv(jc,jd);
covq = jc*cov*jc;
"";"MLE of natural parameters:";;
teta_se = sqrt(diag(covq));
"";"teta_h se";;teta_ho~teta_se;
"";"correlation matrix";;corrvc(covq);


@@@ Posterior results by Gibbs sampling @@@

   @ initialisations @

name = "\204w"~"\204a"~"\204g"~"\204n"; @parameter names@
k = cols(name);                    @number of parameters@
teta_inf = {0.001 0.001 0.001 0.001};@lower bounds of integration ranges@
teta_sup = {1.00   1.00   1.00   30};@upper bounds of integration ranges@
M=33;                              @number of points for grid of integration@
m1 = m;                            @for trapezoidal rule in proc drawi@
w = .5 | reshape(1,m-2,1) | .5;    @weights for trapezoidal rule (see drawi)@
@ m1 = (m+1)/2;@                   @for Simpson in proc drawi@
cdf = zeros(m1,1);                 @see proc drawi@
step = (teta_sup-teta_inf)/(m-1);  @increments@
step2 = 2*step;                    @see proc drawi@
step3 = step2/6;                   @see proc drawi@
teta_gr = zeros(m,k);              @abcissae of each parameter for Simpson@
i=1;
do while i<m+1;
  teta_gr[i,.] = teta_inf + (i-1)*step;
  i=i+1;
endo;
poc = zeros(m,1);                  @to store conditional posterior values@
marg = zeros(m,k);                 @to average conditionals@
f = zeros(m,k);
st = zeros(1,k);
ec = zeros(k,1);
vc = zeros(k,1);
u = zeros(t,1);                    @see proc condi@
v = zeros(t,1);

@load teta_ho = c:\mydoc\articles\gibbs\calculs\teta_ho.fmt;@
TETA_0 = teta_ho;                  @initial conditioning values of Gibbs@
N = 2000;                          @total # draws required@
NDROP = 1000;                      @# left out initial draws for final results@
NS = 1;                            @# intermediate saves in Gibbs loop@
nr = floor(n/ns);                  @# drawings before next save@
n = ns*nr;                         @actual # draws@
teta = zeros(n,k);                 @table of draws for final results@
cexp = zeros(n,k);                 @table of conditional means@
cvar = zeros(n,k);                 @table of conditional variances@

   @ Gibbs loop @

TETA_C = teta_0;
@update teta_c if we are not in first round@
@load previous = c:\mydoc\articles\gibbs\calculs\T1050.fmt;
TETA_C = previous[rows(previous),.]'; @
"teta_c";;teta_c;
datestr(0);;" at ";;timestr(0);

js = 1;
do while js < ns+1;
  j=1;
  jo=(js-1)*nr;
  do while j<nr+1;
    i=1;
    do while i<k+1;
      f[.,i] = condi(m,poc,i,teta_gr[.,i],teta_c,yt,y1,u,v,vy);
      @ f[.,i]: m values of posterior conditional kernel of parameter i @
@      {teta_c[i],st[i]} =@
      {teta_c[i],f[.,i],ec[i],vc[i]} = 
      drawi(m,teta_gr[.,i],f[.,i],step[i],step2[i],step3[i],cdf,i);
      @ update teta_c with draw of parameter i from its conditional posterior
        and gets integral st[i] of conditional posterior in order to normalize
        the conditional posterior f[.,i] values (see 3 lines below) @
      i=i+1;
    endo;        

    if  j < ndrop; goto skipmar; endif;
    marg = marg + f; @adding up of (normalized) conditionals after left out initial draws @
    skipmar:
    teta[jo+j,.] = teta_c';        @assign draw j to table of draws@
    cexp[jo+j,.] = ec';            @assign cond. exp. j to table of draws@
    cvar[jo+j,.] = vc';            @assign cond. var. j to table of draws@
    @"";"j";;j;@
    @"draw";;teta_c';@
    j=j+1;
  endo;

  TIR150AF = teta[1:js*nr,.];
  save path=^spath TIR150AF;
  TES150AF = cexp[1:js*nr,.];
  save path=^spath TES150AF;
  TVA150AF = cvar[1:js*nr,.];
  save path=^spath TVA150AF;
  MAR150AF = marg;
  save path=^spath MAR150AF;
  "";"js-th save done";;js;
  datestr(0);;" at ";;timestr(0);
  js = js+1;
endo;
"last draw";;teta_c;

   @ final results @

pomean = meanc(teta[ndrop+1:n,.]);
povcov = vcx(teta[ndrop+1:n,.]);
pamean = meanc(cexp[ndrop+1:n,.]);
pavcov = meanc(cvar[ndrop+1:n,.]) + diag(vcx(cexp[ndrop+1:n,.]));
"";"Posterior results by Gibbs:";

@ CHOOSE APPROPRIATE STATEMENT ACCORDING TO PRIOR IN PROC POGARCH: @
"Prior on nu: flat on nu";
@"Prior on nu: flat on 1/nu";@
@"Prior on nu: half Cauchy on nu";@

"initial value and integration bounds";;teta_0~teta_inf'~teta_sup';
@ "Simspon rule with rule with ";;m;;" points"; @
"Trapezoidal rule with ";;m;;" points";
nj = n-ndrop;
"number of draws used for results";;nj;
"posterior means and st. dev.";;
pomean~sqrt(diag(povcov))~pamean~sqrt(pavcov);
"posterior correlation matrix";;corrvc(povcov);

marg = marg/nj;
i=1;
do while i<k+1;
  @window(2,2,0);
  setwind(1);
  title("Histogram appr. of post. density of "$+name[i]);
  {b,mid,freq} = histp(teta[ndrop+1:n,i],17);
  setwind(3);@
  title("Posterior density of "$+name[i]);
  xy(teta_gr[.,i],marg[.,i]);
  @endwind;@
  "";"parameter";;i;
  "abs dens";;teta_gr[.,i]~marg[.,i];
   i=i+1;
endo;

end;


proc LFGARCH(teta,x);
@* likelihood evaluation at teta for each data in x
   Student distribution, for MAXLIK
   result is a vector (value of LF for each row of x) *@
local v,nu,nu1h,nuv;
v = recserar(teta[1]^2+teta[2]^2*lag1(uq),vy,teta[3]^2);
nu = teta[4]; nu1h = (nu+1)/2; nuv = nu*v;
retp ( ln(gamma(nu1h))-ln(gamma(nu/2))-0.5*ln(nuv)-nu1h*ln(1+uq./nuv) );
endp;


proc CONDI(m,poc,i,te,te_c,y,x1,u,v,vy);
@* conditional posterior of parameter i evaluated at m values in te,
   conditionally on values of other parameters given in te_c
   result is a vector of m values *@
  local j;
  j=1;
  do while j<m+1;
    te_c[i] = te[j,1];
    poc[j] = pogarch(te_c,y,x1,u,v,vy);
    j=j+1;
  endo;
  retp(poc);
endp;


proc POGARCH(teta,x,x1,u,v,vx);
@* posterior density kernel at teta for data set x
   result is a scalar *@
  local nu,nu1h,nuv,lf;
  v = recserar(teta[1]+teta[2]*lag1(uq),vx,teta[3]);
  nu = teta[4]; nu1h = (nu+1)/2; nuv = nu*v;
  lf = sumc( ln(gamma(nu1h))-ln(gamma(nu/2))
             -0.5*ln(nuv)-nu1h*ln(1+uq./nuv) );
  @ prior is proportional to 1 @
  retp (exp(lf));
  @ prior is proportional to 1/nu^2 @
  @ retp (exp(lf)/nu^2); @
  @ prior is half Cauchy on nu @
  @ retp (exp(lf)/(1+nu^2));@
endp; @ See also lines 224-227 to write the prior on the output @


proc (4) = DRAWI(m,te,pdf,step,step2,step3,cdf,ii);
@* tabulate distribution function cdf at m points given in te, from density
   given in pdf at m1 points, and draws a single realization of the
   random variable by inverting the distribution numerically *@
local som,intg,ec,vc,i,u,cdfp,draw,var;

@ Trapezoidal rule @
som = pdf + lag1(pdf);
som[1] = 0;
cdf = 0.5*step*cumsumc(som);
intg = cdf[m];
pdf = pdf/intg;
som = w.*pdf.*te;
ec = step*sumc(som);
vc = step*sumc(som.*te) - ec^2;

@ draw a uniform number between 0 and cdf[m1] @
u = rndu(1,1)*intg;

@ now find between which values to interpolate @
i = indnv(1,cdf.>u);

@ and interpolate linearly @
cdfp = cdf[i-1];
@draw = te[2*i-3]+(u-cdfp)*step2/(cdf[i]-cdfp);@ @Simpson@
draw = te[i-1] + (u-cdfp)*step/(cdf[i]-cdfp); @Trapeze@
retp (draw,pdf,ec,vc);
@retp (draw,intg);@
endp;


proc STUDT(b,cholmi,nu,k);
@* Draw a Student vector of size k, mean bi and variance nu/(nu-2)*s*M^{-1} 
   NB: cholmi must be equal to chol(M^{-1})' (transpose is crucial) *@
retp( b + cholmi*rndn(k,1)/sqrt(chi2(nu)/nu) );
endp;


proc CHI2(nu);
@* Algorithme dans Rubinstein page 93-94 *@
local k2,k2i;

if nu > 30;
   retp( (rndn(1,1)+sqrt(2*nu-1))^2/2 );
endif;
k2 = nu/2;
k2i = int(nu/2);
if k2 == k2i;
   retp( -2*ln(prodc(rndu(k2,1))) );
else;
   retp( -2*ln(prodc(rndu(k2,1))) + rndn(1,1)^2);
endif;
endp;



/* for DRAWI:
@to check@
"";"i";ii;
"tetai pdf cdf som";;te~pdf~cdf~som;
"sumc(som) step ec vc";;sumc(som)~step~ec~vc;
var = cons;
@window(2,2,0);
setwind(1);@
title("Conditional pdf of "$+name[ii]); xy(te,pdf);
@setwind(3);
title("Conditional cdf of "$+name[ii]); xy(te,cdf);
endwind;@
*/

/* for DRAWI:
@ Simpson's rule @
@local tecdf;@
@tecdf = zeros((m+1)/2,1);@
intg=0;
@tecdf[1] = te[1];@
i=3;
do while i<m+1;
  intg = intg + step3*(pdf[i-2]+4*pdf[i-1]+pdf[i]);
  cdf[(i+1)/2] = intg;
  @tecdf[(i+1)/2] = te[i];@
  i=i+2;
endo;
 
@to check@
"";"i";ii;
"tetai pdf";;te~pdf;
"tecdf cdf";;tecdf~cdf@;
var = cons;
window(2,2,0);
setwind(1);
title("Conditional pdf of "$+name[ii]); xy(te,pdf);
setwind(3);
title("Conditional cdf of "$+name[ii]); xy(tecdf,cdf);
endwind;
*/
